/*
 *	program     : nm.c
 *	author      : d(T.Kawamoto)
 *	Copyright   : 1992 by d(T.Kawamoto)
 *	Function    : symbol lister
 *	Last Modify : 1992/4/12  first release
 *@@@@@@@@@@mifty-rerve     NAH00720    d
 *@@@@@@@@@@d-lail     kawamoto@miln.mei.co.jp
 *@@@@@@@@@@()dVXeÉ
 *@@@@@@@@@@@@@@@@@@@{
 *                1995/05/18  2nd patch
 *                씎Y/oo <HGG00736@niftyserve.or.jp 
 *  version 1.02.1
 *              : 2000/11/11 Takahisa Yokota/KQ <cao01480@pop02.odn.ne.jp>
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#ifndef EOF
#define EOF (-1)
#endif
#define CONTINUE 1

FILE *fptr;
int offset;

/* patched by oo */
int on_memory = 0;
unsigned char *buffer;
int file_size;
const char *program_name;

void
read_seek (int num)
{
  offset += num;
  if (!on_memory)
    {
      fseek (fptr, num, 1);
    }
}

int
read_byte ()
{
  if (on_memory)
    {
      if (offset >= file_size)
        return 0xff;
      return buffer[offset++];
    }
  ++offset;
  return fgetc (fptr) & 0xff;
}

int
read_word ()
{
  int data;

  data = read_byte () << 8;
  data += read_byte ();
  return data;
}

int
read_long ()
{
  int data;

  data = read_byte () << 24;
  data += read_byte () << 16;
  data += read_byte () << 8;
  data += read_byte ();
  return data;
}

void
skip_if_odd ()
{
  if (offset & 1)
    read_byte ();
}

int
label_display ()
{
  int command;

  switch (command = read_byte ())
    {
    case 0x00:
      if (read_byte () == 0x68)
       {
          /* A[JCut@Ci*.lj */
          int obj_size, obj_num, end_offset, rest;
          char *name_buf, *ptr;
          end_offset = offset + 0x18;
          while (offset < end_offset)
            read_byte ();
          read_seek (2 + 4 + 4);
          obj_size = read_long ();
          obj_num = obj_size / 0x30;
          if ((name_buf = malloc (obj_size)) == 0)
            return EOF;
          read_seek (4 + 4);
          rest = read_long () + read_long () - 0x40 - obj_size;
          read_seek (4 + 4);
          ptr = name_buf;
          while (obj_size--)
            *ptr++ = read_byte ();
          while (rest--)
            read_byte ();
          ptr = name_buf;
          while (obj_num--)
           {
             int ofs;
             for (ofs = 0; ofs < 0x18; ++ofs)
               if (ptr[ofs])
                 putchar (ptr[ofs]);
             putchar ('\n');
             ptr += 0x30;
             while (label_display () == CONTINUE);
           }
        }
      return EOF;
    case 0x10:
      {
        int rest_size;

        rest_size = read_byte () + 1;
        while (rest_size--)
         read_byte ();
      }
      skip_if_odd ();
      break;
    case 0x20:
    case 0x30:
      read_seek (1 + 4);
      break;
    case 0x40:
    case 0x41:
    case 0x42:
    case 0x43:
    case 0x45:
    case 0x46:
    case 0x47:
    case 0x80:
      switch (read_byte ())
        {
          case 0xfc:
          case 0xfd:
          case 0xfe:
          case 0xff:
            read_seek (2);
          break;
          case 0x00:
          default:
            read_seek (4);
          break;
        }
      break;
    case 0x48:
      {
        /* st@Ci*.xj */
        int skip_num, read_size, end_offset;
        if (read_byte () != 0x55)
          goto Illegal_format;
        read_seek (2 + 4 + 4);

        /* start address */
        skip_num = read_long ();	/* text segment size */
        skip_num += read_long ();	/* data segment size */
        read_long ();		/* BSS segment size */
        skip_num += read_long ();	/* relocation table size */
        read_size = read_long ();	/* symbol table size */
        read_seek (4 * 8);
        read_seek (skip_num);
        end_offset = offset + read_size;
        while (offset < end_offset)
          {
            int c;
            read_byte ();
            c = read_byte ();
            switch (c)
            {
            /* The symbol value is absolute */
            case 0x00:
              printf ("%08X  A ", read_long ());
            break;
            /* The symbol value is absolute */
           case 0x01:
              printf ("%08X  T ", read_long ());
           break;
           case 0x02:
              printf ("%08X  D ", read_long ());
           break;
            /* The symbol value is the uninitalized data section(known as BSS) */
           case 0x03:
             printf ("%08X  B ", read_long ());
           break;
           case 0x04:
             printf ("%08X  S ", read_long ());
           break;
            case 0x05:
              printf ("%08X RD ", read_long ());
            break;
            case 0x06:
              printf ("%08X RB ", read_long ());
             break;
            case 0x07:
              printf ("%08X RS ", read_long ());
              break;
            case 0x08:
              printf ("%08X LD ", read_long ());
              break;
            case 0x09:
              printf ("%08X LB ", read_long ());
              break;
            case 0x0a:
              printf ("%08X LS ", read_long ());
              break;
            default:
              printf ("%08X ?? ", read_long ());
              break;
            }
            while (c = read_byte ())
              putchar (c);
            putchar ('\n');
            skip_if_odd ();
          }
        }
      return EOF;
    case 0x4c:
    case 0x4d:
      /* .ctor .dtor section */
      switch (read_byte ())
        {
        /* it allow only text section */
        case 0x01:
          read_long ();
          break;
        default:
          printf ("Invaild CTOR or DTOR section \n");
          return EOF;
        }
        break;
    case 0x50:
    case 0x51:
    case 0x52:
    case 0x53:
    case 0x55:
    case 0x56:
    case 0x57:
      switch (read_byte ())
        {
          case 0xFE:
          case 0xFF:
            read_word ();
            break;
          case 0x00:
          default:
            read_long ();
            break;
         }
      read_long ();
      break;
    case 0x65:
    case 0x69:
      read_seek (1 + 4 + 2);
      break;
    case 0x90:
    case 0x91:
    case 0x92:
    case 0x93:
    case 0x96:
    case 0x99:
    case 0xa0:
      read_byte ();
      break;
    case 0xb0:
    case 0xb2:
      {
        int c;
        switch (read_byte ())
        {
      case 0x00:
        printf ("%08X  A ", read_long ());
        break;
      case 0x01:
       /* The symbol is text(code) section */
        printf ("%08X  T ", read_long ());
        break;
      case 0x02:
       /* The symbol value is the initalized data section */
        printf ("%08X  D ", read_long ());
        break;
      case 0x03:
       /* The symbol value is the uninitalized data section(known as BSS) */
        printf ("%08X  B ", read_long ());
        break;
      case 0x04:
       /* The symbol value is the stack section (RSTACK) */
        printf ("%08X  S ", read_long ());
        break;
      case 0x05:
       /* The symbol value is the relocate initalized data section (RDATA) */
        printf ("%08X RD ", read_long ());
        break;
      case 0x06:
      /* The symbol value is the relocate uninitalized data section (RBSS) */
        printf ("%08X RB ", read_long ());
        break;
      case 0x07:
      /* The symbol value is the relocate stack section (RSTACK) */
        printf ("%08X RS ", read_long ());
        break;
      case 0x08:
      /* The symbol value is the remote initalized data section (RLDATA) */
        printf ("%08X LD ", read_long ());
        break;
      case 0x09:
       /* The symbol value is the remote initalized data section (RLBSS) */
        printf ("%08X LB ", read_long ());
        break;
      case 0x0A:
       /* The symbol value is the remote stack section (RLSTACK) */
        printf ("%08X LS ", read_long ());
        break;
      case 0xFC:
        /* The symbol is rcommon */
        printf ("         RC ");
        read_long ();
        break;
      case 0xFD:
        /* The symbol is rlcommon */
        printf ("         LC ");
        read_long ();
        break;
      case 0xFE:
        /* The symbol is common */
        printf ("          C ");
        read_long ();
        break;
      case 0xFF:
        /* The symbol is undefined */
        printf ("          U ");
        read_long ();
        break;
      default:
        read_long ();
        break;
      }
      while (c = read_byte ())
        putchar (c);
        putchar ('\n');
        skip_if_odd ();
      }
      break;
    case 0xC0:
      /* header */
      read_seek (1 + 4);
      while (read_byte ());
      skip_if_odd ();
      break;
    case 0xD0:
      /* object filename */
      {
      /* GNU  nm ̓XLbv܂B */
        int c;
        read_seek (1 + 4);
#if defined (__ORIGINAL__)
        printf ("          F ");
        while (c = read_byte ())
          putchar (c);
        putchar ('\n');
#else
        while (c = read_byte ());
#endif
        skip_if_odd ();
      }
      break;
    case 0xD1:
      {
        /* A[JCut@Ci*.aj */
        int c, ofs;
        read_seek (1 + 4);
        while ((c = read_byte ()) != 0xff)
          {
            putchar (c);
            for (ofs = 1; ofs < 0x18; ++ofs)
              if (c = read_byte ())
                 putchar (c);
            putchar ('\n');
            read_seek (4 + 4);
            while (label_display () == CONTINUE);
         }
      }
      return EOF;
    case 0xE0:
      switch (read_byte ())
       {
         case 0x00:
         /* exec address request */
          read_seek (2 + 4);
          break;
         case 0x0c:
         case 0x0d:
         /* '.doctor' '.dodtor' section */
           break;
         default:
         /* object request */
           while (read_byte ());
            skip_if_odd ();
       }
      break;
    default:

    Illegal_format:
      printf ("Illegal file format\n");
      return EOF;
    }
  return CONTINUE;
}

int
main (int argc, char *argv[])
{
  unsigned long free_memory;
  struct stat st;

  program_name = _basename (argv[0]);
  while (*++argv)
    {
      if (strncmp (*argv, "-V", 2) == 0 || strncmp (*argv, "-version", 8) == 0)
        {
          printf ("%s version 1.02.1\n", program_name);
          puts ("Copyright 1992-95 by d(T.Kawamoto), oo");
          puts ("          2000 KQ");
          exit (0);
        }
      if (!strncmp (*argv, "--help", 6))
        {
          printf ("Usage :%s [-V] [--version] [--help] [file ...]\n",program_name);
          printf ("%s supported targets: m68k-sharp-human executable archive librarian binary\n",program_name);
          exit (0);
        }
      if ((fptr = fopen (*argv, "rb")) == NULL)
        {
          fprintf (stderr, "cannot open file %s\n", *argv);
          continue;
        }
      if (fstat (fileno (fptr), &st) == -1)
        {
          fprintf (stderr, "cannot get stat %s\n", *argv);
          continue;
        }
      offset = 0;
      printf ("%s\n", *argv);

      /* 64KoCgPʂɊۂ߂t[TCY */
      free_memory = (chkml () - 0x10000) & ~0xffff;
      file_size = (int) st.st_size;
      if (file_size < free_memory)
        {
           if ((buffer = malloc (file_size)) == NULL)
            {
              fprintf (stderr, "cannot allocate memory\n");
              continue;
            }
           if (fread (buffer, file_size, 1, fptr) != 1)
            {
              fprintf (stderr, "read error %s\n", *argv);
              continue;
            }
           fclose (fptr);
           on_memory = 1;
           while (label_display () == CONTINUE);
           free (buffer);
        }
      else
        {
          on_memory = 0;
          while (label_display () == CONTINUE);
          fclose (fptr);
        }
    }
  return 0;
}
